use sqlx::{SqlitePool, migrate::Migrator, prelude::FromRow};
use uuid::Uuid;

use crate::script::{Parameter, Script};

static MIGRATOR: Migrator = sqlx::migrate!("./migrations");

// TODO: Add Custom Error Type
#[allow(dead_code)]
pub async fn init_db(pool: &SqlitePool) -> Result<(), String> {
    MIGRATOR.run(pool).await.map_err(|e| e.to_string())?;
    Ok(())
}

pub async fn open<T: AsRef<str>>(path: T) -> Result<SqlitePool, String> {
    let db_url = format!("sqlite://{}", path.as_ref());

    SqlitePool::connect(&db_url)
        .await
        .map_err(|e| e.to_string())
}

pub async fn new<T: AsRef<str>>(path: T) -> Result<SqlitePool, String> {
    std::fs::File::create_new(path.as_ref()).map_err(|e| e.to_string())?;
    let db_url = format!("sqlite://{}", path.as_ref());

    SqlitePool::connect(&db_url)
        .await
        .map_err(|e| e.to_string())
}

pub async fn import_scripts(pool: &SqlitePool, scripts: Vec<Script>) -> Result<(), String> {
    for script in scripts {
        add_script(pool, &script).await?;
    }
    Ok(())
}

// TODO: Write the whole Script into the DB
// TODO: Add Custom Error type
#[allow(dead_code)]
pub async fn add_script(pool: &SqlitePool, script: &Script) -> Result<(), String> {
    let id = Uuid::new_v4().to_string();
    let name = script.name();
    let path = script.path();

    #[allow(clippy::expect_used)]
    let parameters = serde_json::to_string(script.parameters())
        .expect("converting to json string should never fail");

    let default_param_set = script.default_parameter_set();
    let help_uri = script.help_uri();
    let supports_paging = script.supports_paging();
    let supports_should_process = script.supports_should_process();
    let positional_bindings = script.positional_binding();

    sqlx::query!(
        r#"INSERT INTO scripts (id, name, path, parameters, default_parameter_set, help_uri, supports_paging, supports_should_process, positional_binding) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)"#,
        id,
        name,
        path,
        parameters,
        default_param_set,
        help_uri,
        supports_paging,
        supports_should_process,
        positional_bindings,
    )
    .execute(pool)
    .await
    .map_err(|e| e.to_string())?;

    Ok(())
}

#[derive(FromRow, Debug)]
struct ScriptRow {
    id: String,
    name: String,
    path: String,
    parameters: String,
    default_parameter_set: Option<String>,
    help_uri: Option<String>,
    supports_paging: Option<String>,
    supports_should_process: Option<bool>,
    positional_binding: Option<bool>,
}

#[allow(dead_code)]
pub async fn get_script<T: AsRef<str>>(pool: &SqlitePool, id: T) -> Result<(Uuid, Script), String> {
    let script_row =    sqlx::query_as::<_, ScriptRow>("SELECT id, name, path, parameters, default_parameter_set, help_uri, supports_paging, supports_should_process, positional_binding FROM scripts WHERE id = ?").bind(id.as_ref())
        .fetch_one(pool)
        .await
        .map_err(|e| e.to_string())?;
    let params: Vec<Parameter> =
        serde_json::from_str(&script_row.parameters).map_err(|e| e.to_string())?;
    Ok((
        Uuid::parse_str(&script_row.id).map_err(|e| e.to_string())?,
        Script::builder()
            .set_name(script_row.name)
            .set_path(script_row.path)
            .set_parameters(params)
            .set_default_parameter_set(script_row.default_parameter_set)
            .set_help_uri(script_row.help_uri)
            .set_supports_paging(script_row.supports_paging)
            .set_supports_should_process(script_row.supports_should_process)
            .set_positional_binding(script_row.positional_binding)
            .build()
            .map_err(|e| e.to_string())?,
    ))
}

// TODO: Return Vec<Script> instead of String-Tuple
// TODO: Add Custom Error Type
#[allow(dead_code)]
pub async fn list_scripts(pool: &SqlitePool) -> Result<Vec<(Uuid, Script)>, String> {
    sqlx::query_as::<_, ScriptRow>("SELECT id, name, path, parameters, default_parameter_set, help_uri, supports_paging, supports_should_process, positional_binding FROM scripts")
        .fetch_all(pool)
        .await
        .map_err(|e| e.to_string())?
        .into_iter()
        .map(|row| {
            let params: Vec<Parameter> =
                serde_json::from_str(&row.parameters).map_err(|e| e.to_string())?;
            Ok((
                Uuid::parse_str(&row.id).map_err(|e| e.to_string())?,
                Script::builder()
                    .set_name(row.name)
                    .set_path(row.path)
                    .set_parameters(params)
                    .set_default_parameter_set(row.default_parameter_set)
                    .set_help_uri(row.help_uri)
                    .set_supports_paging(row.supports_paging)
                    .set_supports_should_process(row.supports_should_process)
                    .set_positional_binding(row.positional_binding)
                    .build()
                    .map_err(|e| e.to_string())?,
            ))
        })
        .collect()
}

#[cfg(test)]
mod tests {
    #![allow(unused_imports, clippy::unwrap_used, clippy::expect_used)]
    use super::*;

    #[tokio::test]
    async fn test_insert_and_list() {
        let db_name = Uuid::new_v4().to_string();
        let db_path = format!("./{db_name}");
        let db_url = format!("sqlite://{db_name}");

        std::fs::File::create_new(&db_path).unwrap();

        let pool = SqlitePool::connect(&db_url).await.unwrap();

        assert!(init_db(&pool).await.is_ok());

        let script = Script::from_file("powershell/test-script.ps1").unwrap();

        assert!(add_script(&pool, &script).await.is_ok());

        let got = list_scripts(&pool).await.unwrap();

        assert_eq!(got.len(), 1);
        assert_eq!(got[0].1.name(), "test-script.ps1");
        assert_eq!(got[0].1.parameters().len(), 9);
        std::fs::remove_file(&db_path).unwrap();
    }
}
