use std::{fmt::Display, str::FromStr};

use serde::{Deserialize, Serialize};

#[derive(Serialize, Clone, Debug)]
pub enum DataType {
    DateTime,
    String,
    Int32,
    Int64,
    #[serde(rename = "SwitchParameter")]
    Switch,
    Boolean,
    PSCredential,
    SecureString,
}

impl Display for DataType {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str(match self {
            Self::DateTime => "DateTime",
            Self::String => "String",
            Self::Int32 => "Int32",
            Self::Int64 => "Int64",
            Self::Switch => "SwitchParameter",
            Self::Boolean => "Boolean",
            Self::PSCredential => "PSCredential",
            Self::SecureString => "SecureString",
        })
    }
}

impl FromStr for DataType {
    //TODO: Create Custom Error Type
    type Err = String;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let s = s.to_uppercase();
        match s.as_str() {
            "STRING" => Ok(Self::String),
            "DATETIME" => Ok(Self::DateTime),
            "INT32" => Ok(Self::Int32),
            "INT64" => Ok(Self::Int64),
            "SWITCHPARAMETER" => Ok(Self::Switch),
            "BOOLEAN" => Ok(Self::Boolean),
            "PSCREDENTIAL" => Ok(Self::PSCredential),
            "SECURESTRING" => Ok(Self::SecureString),
            _ => Err(String::from("Invalid Datatype")),
        }
    }
}

impl<'de> Deserialize<'de> for DataType {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: serde::Deserializer<'de>,
    {
        let s = String::deserialize(deserializer)?;
        DataType::from_str(&s)
            .map_err(|_| serde::de::Error::custom(format!("invalid Datatype: {}", s)))
    }
}
